import os
from .app import (
    is_process_running,
    kill_process,
    kill_service,
    start_process,
    start_service,
)
from .input import cinput
from .print import cprint


class Surfshark:
    def __init__(self, path: str):
        self.__path = path

    def running(self):
        return is_process_running("Surfshark.exe")

    def ensure(self):
        """
        Ensures Surfshark is running and pauses execution if not.
        """
        if not self.running():
            cprint("<y>Surfshark is not running. Starting Surfshark now...")
            self.start()  # Ensure
        else:
            print("Surfshark is running. Proceeding...")

        # Surfshark may take a while to start up and connect.
        while not self.running():
            cinput("Press <y><ENTER><w> once Surfshark has connected to continue...")

    def kill(self):
        kill_process("Surfshark.exe")
        kill_service("Surfshark Service")
        kill_service("Surfshark WireGuard Service")

    def start(self):
        # Main application
        start_process(self.__path)
        # Background service
        start_service("Surfshark Service")
        start_service("Surfshark WireGuard Service")



# Third-party libraries
from gspread_formatting import CellFormat

# Local libraries
from lytils.file import write_to_file
from .format import DefaultFormat


class Column:
    """
    Used for defining columns and how they're to be formatted.

    Args:
        id (str): The unique identifier for the column. Should be snake_case.
        header (str): The header text for the column.
        format (CellFormat, optional): The format to apply to the column. Defaults to DefaultFormat().
        width (int, optional): The width of the column. Defaults to 100.
    """

    def __init__(
        self,
        id: str,
        header: str,
        format: CellFormat = DefaultFormat(),
        width: int = 100,
    ):
        self.__id = id
        self.__header = header
        self.__format = format
        self.__width = width

    def get_id(self):
        return self.__id

    def get_header(self):
        return self.__header

    def get_format(self):
        return self.__format

    def get_width(self):
        return self.__width


class Columns:
    """
    Used for grouping Column objects.

    Args:
        columns (list): A list of Column objects.
    """

    def __init__(self, columns: list[Column]):
        self.__columns = {column.get_id(): column for column in columns}
        self.__order = [column.get_id() for column in columns]

    def output_item_class(self, output_path: str):
        """
        Generates a Python class definition for a data structure with attributes
        based on the columns specified in `self.__order` and writes it to a file.

        The generated class is a dataclass named `CustomItem` with each column
        in `self.__order` as a string attribute initialized to an empty string.

        Args:
            output_path (str): The file path where the generated class definition
                               will be written.

        Example:
            If `self.__order` is ['id', 'name', 'age'], the generated class will be:

            from dataclasses import dataclass

            @dataclass
            class CustomItem:
                id: str = ""
                name: str = ""
                age: str = ""

        This class definition will be written to the file specified by `output_path`.
        Item values default to an empty string, but can be changed as needed.
        """
        file_str = "from dataclasses import dataclass\n\n"
        file_str = "".join(
            [
                "from dataclasses import dataclass\n",
                "\n",
                "@dataclass\n",
                "class CustomItem:\n",
                *[f'    {column}: str = ""\n' for column in self.__order],
                "\n",
            ]
        )
        write_to_file(output_path, file_str)



import logging
from datetime import datetime, timezone, timedelta
from inspect import currentframe
from lytils import cprint, cstrip, ctext
from lytils.file import LyFile


def get_line_number():
    cf = currentframe()
    return cf.f_back.f_lineno


class InvalidLogLevelException(Exception):
    # Raise this when undetected_chromedriver is not installed
    def __init__(
        self,
        message=f"<y>Parameter 'level' must equal one of the following: debug, info, warning, error, critical.",
    ):
        self.message = ctext(message)
        super().__init__(self.message)


class LyLoggerNotSetUpException(Exception):
    # Raise this when undetected_chromedriver is not installed
    def __init__(
        self,
        message=f"<y>LyLogger.setup() not ran.",
    ):
        self.message = ctext(message)
        super().__init__(self.message)


# Custom filter class to allow only a specific log level
class LevelFilter(logging.Filter):
    def __init__(self, level):
        super().__init__()
        self.level = level

    def filter(self, record):
        # Allow only log messages of the specified level
        return record.levelno == self.level


# Custom Formatter to Use a Specific Timezone Offset
class TimezoneFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None, timezone_offset=0):
        super().__init__(fmt, datefmt)
        # Set the desired timezone using the offset provided
        self.timezone = timezone(timedelta(hours=timezone_offset))

    def formatTime(self, record, datefmt=None):
        # Convert the record's creation time to datetime and apply the timezone
        dt = datetime.fromtimestamp(record.created, tz=self.timezone)
        # Format the time according to the datefmt if specified, otherwise default to ISO format
        if datefmt:
            s = dt.strftime(datefmt)
        else:
            s = dt.isoformat()
        return s


class LyLogger:
    LEVELS = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }

    def __init__(
        self,
        name: str,
        path: str = "",
        persistent: bool = False,
        encoding: str = "utf-8",
        formatting: str = "%(asctime)s - [%(levelname)s] %(message)s",
        date_format: str = "%Y-%m-%d %H:%M:%S",
        local_time: bool = True,
        timezone_offset: int = 0,
        level="info",
        debug_path: str = "",
        info_path: str = "",
        warning_path: str = "",
        error_path: str = "",
        critical_path: str = "",
        use_logger: bool = True,
    ):
        """
        Wrapper around the 'logging' package.

        Args:
            name (str): Each instance of logging needs a name to associate its handlers.
            path (str): The default path of the log file(s). Defaults to 'logs/{name}.log'.
            persistent (bool): Set to True if log file(s) should not be overwritten.
            encoding (str): The encoding of the log file(s).
            formatting (str): The formatting of the prefix for each line in the log file(s).
            date_format (str): The date format of the timestamp for each line in the log file(s).
            level (str): The importance level of messages to be shown. Least to most: debug > info > warning > error > critical.
            debug_path (str): The path of the debug logs (ex. file.debug.log). Defaults to path.
            info_path (str): The path of the info logs (ex. file.info.log). Defaults to path.
            warning_path (str): The path of the warning logs (ex. file.warning.log). Defaults to path.
            error_path (str): The path of the error logs (ex. file.error.log). Defaults to path.
            critical_path (str): The path of the critical logs (ex. file.critical.log). Defaults to path.
            use_logger (bool): Set to False if you don't wish to actually create any logs.
                The reason for this is to be able to pass use_logger at initializing LyLogger instead of
                having to constantly make checks in the calling code if logs are desired.
        """
        self.use_logger = use_logger

        if name == "":
            raise ValueError("Logger name cannot be empty.")

        self.name = name
        self.path = f"{path}/{name}.log" if path else f"logs/{name}.log"
        self.encoding = encoding
        self.formatting = formatting
        self.date_format = date_format
        self.local_time = local_time
        self.timezone_offset = timezone_offset

        if level not in LyLogger.LEVELS.keys():
            raise InvalidLogLevelException

        self.level = LyLogger.LEVELS[level]

        # Allow for multiple log files
        self.debug_path = debug_path if debug_path else self.path
        self.info_path = info_path if info_path else self.path
        self.warning_path = warning_path if warning_path else self.path
        self.error_path = error_path if error_path else self.path
        self.critical_path = critical_path if critical_path else self.path

        self._logger = None
        self.setup(persistent)

    def _get_file_handler(self, path: str, level, formatter):
        file_handler = logging.FileHandler(path)
        file_handler.setLevel(LyLogger.LEVELS[level])
        file_handler.addFilter(LevelFilter(LyLogger.LEVELS[level]))
        file_handler.setFormatter(formatter)
        return file_handler

    def setup(self, persistent: bool = False) -> None:
        """
        Set up individual instance of logging.

        Args:
            persistent (bool): Set to True if log file should not be overwritten.

        """
        if not self.use_logger:
            return

        # Create new log files if they should be overwritten or if it doesn't exist.
        for path in {
            self.path,
            self.debug_path,
            self.info_path,
            self.warning_path,
            self.error_path,
            self.critical_path,
        }:
            log_file = LyFile(path)
            if not persistent or not log_file.exists():
                log_file.create()

        logger = logging.getLogger(self.name)
        logger.setLevel(self.level)

        formatter = (
            logging.Formatter(self.formatting, self.date_format)
            if self.local_time
            else TimezoneFormatter(
                self.formatting, self.date_format, self.timezone_offset
            )
        )

        # Get and add file handlers
        debug_handler = self._get_file_handler(self.debug_path, "debug", formatter)
        logger.addHandler(debug_handler)

        info_handler = self._get_file_handler(self.info_path, "info", formatter)
        logger.addHandler(info_handler)

        warning_handler = self._get_file_handler(
            self.warning_path, "warning", formatter
        )
        logger.addHandler(warning_handler)

        error_handler = self._get_file_handler(self.error_path, "error", formatter)
        logger.addHandler(error_handler)

        critical_handler = self._get_file_handler(
            self.critical_path, "critical", formatter
        )
        logger.addHandler(critical_handler)

        self._logger = logger

    # * GETTERS

    def get_debug_path(self) -> str:
        """
        Get the path of the debug log file.

        Returns:
            str: The path of the debug log file.
        """
        return self.debug_path

    def get_info_path(self) -> str:
        """
        Get the path of the info log file.

        Returns:
            str: The path of the info log file.
        """
        return self.info_path

    def get_warning_path(self) -> str:
        """
        Get the path of the warning log file.

        Returns:
            str: The path of the warning log file.
        """
        return self.warning_path

    def get_error_path(self) -> str:
        """
        Get the path of the error log file.

        Returns:
            str: The path of the error log file.
        """
        return self.error_path

    def get_critical_path(self) -> str:
        """
        Get the path of the critical log file.

        Returns:
            str: The path of the critical log file.
        """
        return self.critical_path

    # * REGULAR LOG FUNCTIONS

    def debug(self, message: str):
        """
        Logs message of level: debug.

        Args:
            message (str): Message to log.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            self._logger.debug(message)

    def info(self, message: str):
        """
        Logs message of level: info.

        Args:
            message (str): Message to log.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            self._logger.info(message)

    def warning(self, message: str):
        """
        Logs message of level: warning.

        Args:
            message (str): Message to log.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            self._logger.warning(message)

    def error(self, message: str):
        """
        Logs message of level: error.

        Args:
            message (str): Message to log.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            self._logger.error(message)

    def critical(self, message: str):
        """
        Logs message of level: critical.

        Args:
            message (str): Message to log.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            self._logger.critical(message)

    # * JSON LOG FUNCTIONS

    def debug_json(self, message: str, data, indent: int = 4):
        """
        Logs message and json of level: debug.

        Args:
            message (str): Message to log.
            data: Json data to log.
            indent (int): Indentation of json data.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            self._logger.debug(message)
            file = LyFile(self.debug_path)
            file.append_json(data=data, indent=indent)

    def info_json(self, message: str, data, indent: int = 4):
        """
        Logs message and json of level: info.

        Args:
            message (str): Message to log.
            data: Json data to log.
            indent (int): Indentation of json data.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            self._logger.info(message)
            file = LyFile(self.info_path)
            file.append_json(data=data, indent=indent)

    def warning_json(self, message: str, data, indent: int = 4):
        """
        Logs message and json of level: warning.

        Args:
            message (str): Message to log.
            data: Json data to log.
            indent (int): Indentation of json data.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            self._logger.warning(message)
            file = LyFile(self.warning_path)
            file.append_json(data=data, indent=indent)

    def error_json(self, message: str, data, indent: int = 4):
        """
        Logs message and json of level: error.

        Args:
            message (str): Message to log.
            data: Json data to log.
            indent (int): Indentation of json data.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            self._logger.error(message)
            file = LyFile(self.error_path)
            file.append_json(data=data, indent=indent)

    def critical_json(self, message: str, data, indent: int = 4):
        """
        Logs message and json of level: critical.

        Args:
            message (str): Message to log.
            data: Json data to log.
            indent (int): Indentation of json data.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            self._logger.critical(message)
            file = LyFile(self.critical_path)
            file.append_json(data=data, indent=indent)

    # * CPRINT LOG FUNCTIONS

    def cdebug(self, message: str):
        """
        Logs message of level: debug and cprints to terminal.

        Args:
            message (str): Message to log.
        """
        if not self._logger:
            raise LyLoggerNotSetUpException

        if self.use_logger:
            cprint(message)
            text = cstrip(message)
            self._logger.debug(text)

    def cinfo(self, message: str):
        """
        Logs message of level: info and cprints to terminal.

        Args:
            message (str): Message to log.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            cprint(message)
            text = cstrip(message)
            self._logger.info(text)

    def cwarning(self, message: str):
        """
        Logs message of level: warning and cprints to terminal.

        Args:
            message (str): Message to log.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            cprint(message)
            text = cstrip(message)
            self._logger.warning(text)

    def cerror(self, message: str):
        """
        Logs message of level: error and cprints to terminal.

        Args:
            message (str): Message to log.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            cprint(message)
            text = cstrip(message)
            self._logger.error(text)

    def ccritical(self, message: str):
        """
        Logs message of level: critical and cprints to terminal.

        Args:
            message (str): Message to log.
        """
        if self.use_logger:
            if not self._logger:
                raise LyLoggerNotSetUpException
            cprint(message)
            text = cstrip(message)
            self._logger.critical(text)



# Standard Libraries
import time

try:
    import msvcrt
except ModuleNotFoundError:
    msvcrt = None

# Local Libraries
from .print import cprint, ctext


def cinput(text, level=0):
    str_level = ""
    for _ in range(0, level):
        str_level += "  "

    return input(f"{str_level}{ctext(text)}")


def pause(text="Press <y><ENTER><w> to continue..."):
    return cinput(text)


def press_enter_or_wait_to_continue(seconds):
    t0 = time.time()
    cprint(f"Press <y><ENTER><w> or <y>wait {seconds} seconds<w>...")
    while time.time() - t0 < 300:
        if msvcrt.kbhit():
            if msvcrt.getch() == b"\r":  # not '\n'
                break
        time.sleep(0.1)



from dataclasses import dataclass, fields


@dataclass
class Item:
    def to_dict(self):
        """
        Converts the dataclass attributes to a dictionary.

        Returns:
            dict: A dictionary with attribute names as keys and their values.
        """
        return {field.name: getattr(self, field.name) for field in fields(self)}



import pandas as pd
from lytils.regex import match, replace


def convert_accounting_to_numeric(column: pd.Series):
    def transform_column(text: str) -> float:
        text = replace(r"[\$,-]", "", text).strip()
        return float(text) if text != "" else 0

    return column.apply(transform_column)


def get_column_letter(column_number):
    """
    Get column letter given column number.
    1 = A, 2 = B, ...,  27 = AA, 28 = AB, etc
    """
    dividend = column_number
    letter = ""
    while dividend > 0:
        remainder = (dividend - 1) % 26
        letter = chr(65 + remainder) + letter
        dividend = (dividend - remainder) // 26
    return letter


def get_column_number(column_letter):
    """
    Get column number given column letter.
    A = 1, B = 2, ...,  AA = 27, AB = 28, etc
    """
    number = 0
    for char in column_letter:
        number = number * 26 + (ord(char) - ord("A") + 1)
    return number


def get_header_range(range: str):
    """
    Get header range (first row) given a range in a1 notation
    """
    first_column = match(r"[A-Z]+", range, group=1)
    last_column = match(r"[A-Z]+", range, group=2)
    first_row = match(r"[0-9]+", range, group=1)

    return f"{first_column}{first_row}:{last_column}{first_row}"


# Get data range from a range in a1 notation
def get_data_range(range: str):
    """
    Get data range (everything except first row) given a range in a1 notation
    """
    first_column = match(r"[A-Z]+", range, group=1)
    first_row = match(r"[0-9]+", range, group=1)
    last_column = match(r"[A-Z]+", range, group=2)
    last_row = match(r"[0-9]+", range, group=2)

    # Increment first row by 1
    return f"{first_column}{int(first_row) + 1}:{last_column}{last_row}"



import traceback

COLORS = {
    "white": "\033[0m",  # default,
    "red": "\033[31m",
    "green": "\033[32m",
    "yellow": "\033[33m",
    "blue": "\033[34m",
    "magenta": "\033[35m",
    "cyan": "\033[36m",
}


def ctext(text):
    text = text.replace("<w>", COLORS["white"])
    text = text.replace("<r>", COLORS["red"])
    text = text.replace("<g>", COLORS["green"])
    text = text.replace("<y>", COLORS["yellow"])
    text = text.replace("<b>", COLORS["blue"])
    text = text.replace("<m>", COLORS["magenta"])
    text = text.replace("<c>", COLORS["cyan"])

    return f"{text}{COLORS['white']}"


def cstrip(text):
    """
    Strips color tags from text.
    """
    text = text.replace("<w>", "")
    text = text.replace("<r>", "")
    text = text.replace("<g>", "")
    text = text.replace("<y>", "")
    text = text.replace("<b>", "")
    text = text.replace("<m>", "")
    text = text.replace("<c>", "")

    return f"{text}{COLORS['white']}"


def cprint(text, level=0):
    text = ctext(text)

    str_level = ""
    for i in range(0, level):
        str_level += "  "

    print(f"{str_level}{text}{COLORS['white']}")


def print_trace(color="red"):
    set_print_color(color)
    traceback.print_exc()
    reset_print_color()


def reset_print_color():
    print(COLORS["white"])


def set_print_color(color):
    print(COLORS[color])



import ast
import inspect
import importlib
from lytils import ctext


class MissingImportsException(Exception):
    # Raise this when undetected_chromedriver is not installed
    def __init__(self, message: str):
        self.message = ctext(f"<y>{message}")
        super().__init__(self.message)


def get_imports(filename):
    with open(filename, "r") as file:
        code = file.read()

    tree = ast.parse(code)
    imports = [
        node.module if isinstance(node, ast.ImportFrom) else node.names[0].name
        for node in ast.walk(tree)
        if isinstance(node, (ast.Import, ast.ImportFrom))
    ]
    return imports


def raise_missing_imports(missing_imports, method: str = "poetry"):
    """
    Raise an exception if there are missing imports.

    Args:
        method (str, optional): The method to install missing imports. Defaults to "poetry". Can also pass "pip".
    """
    if missing_imports:
        prefix = ""
        if method == "poetry":
            prefix = "poetry add "
        elif method == "pip":
            prefix = "pip install "
        message = 'Try running "' + prefix + " ".join(missing_imports) + '"'
        raise MissingImportsException(message)


def verify_imports(filename: str = "main.py", method: str = "poetry"):
    """
    Verify that all imports are installed.

    Args:
        filename (str, optional): The filename to verify imports. Defaults to "main.py".
        method (str, optional): The method to install missing imports. Defaults to "poetry". Can also pass "pip".
    """
    imports = get_imports(filename)
    missing_imports = []
    for module in imports:
        try:
            importlib.import_module(module)
        except ImportError:
            missing_imports.append(module)

    raise_missing_imports(missing_imports, method)



try:
    import requests

    REQUESTS_IMPORTED = True
except ModuleNotFoundError:
    REQUESTS_IMPORTED = True

if REQUESTS_IMPORTED:

    def get_response_dict(response: requests.Response):
        # Try to get the JSON content
        try:
            json_content = response.json()
        except ValueError:
            json_content = ""

        try:
            content = response.content.decode(response.encoding)
        except UnicodeDecodeError:
            # Fallback to latin1 if UTF-8 and apparent_encoding fail
            content = response.content.decode("latin1")

        # Create a dictionary from the response object
        return {
            "status_code": response.status_code,
            "text": response.text,
            "content": content,  # Decoding bytes to string
            "json": json_content,
            "headers": dict(response.headers),
            "url": response.url,
            "encoding": response.encoding,
            "elapsed": response.elapsed.total_seconds(),  # Converting timedelta to seconds
            "cookies": requests.utils.dict_from_cookiejar(response.cookies),
            "history": [resp.url for resp in response.history],
            "reason": response.reason,
        }

    def save_response_html(response: requests.Response, path: str):
        with open(path, "w", encoding=response.encoding) as file:
            file.write(response.text)

else:

    def get_response_dict(response):
        raise ModuleNotFoundError("requests module is not installed.")

    def save_response_html(response, path: str):
        raise ModuleNotFoundError("requests module is not installed.")



import gspread as gs
from .SpreadsheetTab import SpreadsheetTab


class Spreadsheet:
    def __init__(self, sheet: gs.Spreadsheet):
        self.__sheet = sheet

    def get_tab(self, name: str) -> SpreadsheetTab:
        tab = self.__sheet.worksheet(name)
        return SpreadsheetTab(tab=tab)



# Third-party Libraries
import gspread as gs

# Local Libraries
from lytils.google_sheets.Spreadsheet import Spreadsheet


# src: https://practicaldatascience.co.uk/data-science/how-to-read-google-sheets-data-in-pandas-with-gspread
class GoogleSheets:
    """
    Creates a GoogleSheets object that allows for interfacing with the Google Sheets API.
    Using this GoogleSheets object, you can access any sheet by its URL, assuming the service account has permissions.

    Rough steps to get started:
    1. Create a Google Cloud Platform project
    2. Enable the Google Sheets API
    3. Create a service account and add a JSON key to it
    4. Download the JSON key and move it to your project directory
    5. Share any desired sheet with the service account email
    6. Initialize a GoogleSheets object with the path to the service account JSON
    7. Plug the desire sheet url into the get_sheet method
    """

    def __init__(self, service_account: str):
        # Authenticate with Google Sheets
        self.__service_account = gs.service_account(filename=service_account)

    def get_sheet(self, url) -> Spreadsheet:
        sheet = self.__service_account.open_by_url(url)
        return Spreadsheet(sheet=sheet)



from lytils.google_sheets.Columns import Columns
from lytils.google_sheets.SpreadsheetTab import SpreadsheetTab
from lytils.google_sheets.helpers import get_data_range, get_header_range


class Table:
    def __init__(self, tab: SpreadsheetTab, range: str, columns: Columns = None):
        self.__tab = tab
        self.__range = range
        self.__columns = columns

    def setup(self):
        header_range = get_header_range(self.__range)
        data_range = get_data_range(self.__range)
        pass



from gspread_formatting import CellFormat
from gspread_formatting import NumberFormat as GSNumberFormat
from gspread_formatting import TextFormat as GSTextFormat


def DefaultFormat() -> CellFormat:
    return CellFormat(
        numberFormat=GSNumberFormat(type="NUMBER", pattern="General"),
        verticalAlignment="TOP",
    )


def HeaderFormat(**kwargs) -> CellFormat:
    return merge_cell_formats(
        [
            DefaultFormat(),
            CellFormat(
                textFormat=GSTextFormat(bold=True),
                **kwargs,
            ),
        ]
    )


def DateTimeFormat(pattern="yyyy-mm-dd hh:mm:ss", **kwargs) -> CellFormat:
    return merge_cell_formats(
        [
            DefaultFormat(),
            CellFormat(
                numberFormat=GSNumberFormat(type="DATE_TIME", pattern=pattern),
                **kwargs,
            ),
        ]
    )


def TimeFormat(pattern: str = "hh:mm:ss", **kwargs) -> CellFormat:
    return merge_cell_formats(
        [
            DefaultFormat(),
            CellFormat(
                numberFormat=GSNumberFormat(type="TIME", pattern=pattern),
                **kwargs,
            ),
        ]
    )


def AccountingFormat(**kwargs) -> CellFormat:
    return merge_cell_formats(
        [
            DefaultFormat(),
            CellFormat(
                numberFormat=GSNumberFormat(
                    type="NUMBER",
                    pattern='_("$"* #,##0.00_);_("$"* \(#,##0.00\);_("$"* "-"??_);_(@_)',
                ),
                **kwargs,
            ),
        ]
    )


def NumberFormat(pattern="0", **kwargs) -> CellFormat:
    # Possible patterns:
    # "0" -> 1234
    # "0.00" -> 1234.00
    # "#,##0" -> 1,234
    # "#,##0.00" -> 1,234.00
    return merge_cell_formats(
        [
            DefaultFormat(),
            CellFormat(
                numberFormat=GSNumberFormat(type="NUMBER", pattern=pattern),
                **kwargs,
            ),
        ]
    )


def PercentFormat(pattern="0.00%", **kwargs) -> CellFormat:
    return merge_cell_formats(
        [
            DefaultFormat(),
            CellFormat(
                numberFormat=GSNumberFormat(type="PERCENT", pattern=pattern),
                **kwargs,
            ),
        ]
    )


def TextFormat(**kwargs) -> CellFormat:
    return merge_cell_formats(
        [
            DefaultFormat(),
            CellFormat(
                numberFormat=GSNumberFormat(type="TEXT", pattern="@"),
                horizontalAlignment="LEFT",
                **kwargs,
            ),
        ]
    )


def WrapTextFormat(**kwargs) -> CellFormat:
    return merge_cell_formats(
        [
            DefaultFormat(),
            CellFormat(
                numberFormat=GSNumberFormat(type="TEXT", pattern="@"),
                horizontalAlignment="LEFT",
                wrapStrategy="WRAP",
                **kwargs,
            ),
        ]
    )


def merge_cell_formats(formats) -> CellFormat:
    """Merges a list of CellFormat objects into a single CellFormat object.

    Args:
        formats: A list of CellFormat objects.
        [format1, format2, format3]: In the case of similar properties,
            format2 will take priority over format1, and format3 will take
            priority over format1 and format2.

    Returns:
        A CellFormat object that contains the properties of all of the input objects.
    """
    merged_format = CellFormat()
    for format in formats:
        for property_name, property_value in format.__dict__.items():
            if not hasattr(merged_format, property_name):
                setattr(merged_format, property_name, property_value)
            elif property_value != merged_format.__dict__[property_name]:
                setattr(merged_format, property_name, property_value)

    return merged_format



import csv
import json
import os

try:
    from bs4 import BeautifulSoup
except ImportError:
    BeautifulSoup = None


def get_last_id_in_csv_file(file_name, column="id"):
    with open(file_name, "r") as file:
        reader = csv.DictReader(file)

        # Iterate over each row in the CSV file
        for row in reader:
            # Assign the last row to the last_row variable
            last_row = row

        try:
            # Get the value you want from the last row by column name
            desired_value = last_row[column]

            return int(desired_value)
        except:
            return -1


def load_json_from_file(path):
    """
    Load json object from file.
    """
    with open(path, "r") as file:
        return json.load(file)


def create_file_at_path(path: str):
    """
    Creates a blank file at path.
    """
    # Split path into directory and filename
    directory, _ = os.path.split(path)

    # If directory was included, create it if it doesn't exist
    if directory and not os.path.exists(directory):
        os.makedirs(directory)

    open(path, "w").close()  # Create file


def write_to_file(path: str, text: str, encoding: str = "utf-8"):
    """
    Write text to file at path.
    """
    create_file_at_path(path)

    with open(path, "w", encoding=encoding) as file:
        file.write(f"{text}")


def write_html_to_file(path: str, html, encoding: str = "utf-8"):
    """
    Write html to file.
    """
    create_file_at_path(path)

    soup = BeautifulSoup(html, "html.parser")
    formatted_html = soup.prettify()
    with open(path, "w", encoding=encoding) as file:
        file.write(formatted_html)


def write_json_to_file(path: str, data, encoding: str = "utf-8", indent: int = 4):
    """
    Write json object to file.
    """
    create_file_at_path(path)

    with open(path, "w", encoding=encoding) as file:
        json.dump(data, file, indent=indent)


class LyFile:
    def __init__(self, path: str = "LyFile/file.txt"):
        self._path = path

    def exists(self):
        return os.path.exists(self._path)

    def create(self):
        create_file_at_path(self._path)

    def append(self, text: str, encoding: str = "utf-8"):
        """
        Appends text to file at path.
        """
        with open(self._path, "a", encoding=encoding) as file:
            file.write(f"{text}\n")

    def append_json(self, data: dict, encoding: str = "utf-8", indent: int = 4):
        """
        Appends json data to file at path.
        """
        with open(self._path, "a", encoding=encoding) as file:
            json.dump(data, file, indent=indent)
            file.write("\n")



import os
from lytils.file import create_file_at_path

try:
    from airium import Airium
except ImportError:
    Airium = None


def join_styles(styles: dict):
    # Quotes need to be manually added as some styles don't require quotes
    # ex. colors, sizes, etc
    styles_list = []
    for s in styles:
        styles_list.append(f"{s}: {styles[s]};")
    return " ".join(styles_list) if styles_list else ""


def join_attrs(attrs: dict):
    attrs_list = []
    for a in attrs:
        attrs_list.append(f"{a}='{attrs[a]}'")
    return " ".join(attrs_list) if attrs_list else ""


class LogHTML:
    def __init__(self, path: str = "log.html", title: str = "", dark: bool = True):
        self.filename = path
        self.title = title if title else os.path.basename(path)
        self.dark = dark
        self.log = []
        self.paths = {}
        self.__a = Airium()
        self.__a("<!DOCTYPE html>")

    def append_divider(self):
        self.log.append("<hr>")

    def append_image(self, src: str, inline: bool = False):
        display = "inline" if inline else "block"
        self.log.append(f"<img src='{src}' style='display: {display};' />")

    def append_link(self, href: str, text: str = ""):
        text = text if text else href
        attrs = join_attrs({"href": href, "target": "_blank"})
        self.log.append(f"<a {attrs}>{text}</a>")

    def append_text(self, text: str):
        self.log.append(f"<p>{text}</p>")

    def generate_unique_path(self, path: str, name: str, ext: str):
        """
        Generate a unique file path by appending an index to the filename.

        Args:
            path (str): The directory where the file will be saved.
            name (str): The name of the file without the extension.
            ext (str): The file extension without the period.

        Returns:
            str: A unique file path with an appended index if necessary.
        """
        hash = f"{path}/{name}.{ext}"
        hash = hash.replace("//", "/")  # Account for any double slashes
        if hash not in self.paths:
            self.paths[hash] = 0
            return f"{path}/{name}_{0}.{ext}"
        else:
            self.paths[hash] += 1
            return f"{path}/{name}_{self.paths[hash]}.{ext}"

    def style_css(self):
        return """
            body.dark {
                background-color: #333333;
                color: #e0e0e0;

                a {
                    color: #66ccff
                }

                a:visited {
                    color: #bb00cc
                }
            }
            """

    def save(self):
        theme = "dark" if self.dark else "light"
        with self.__a.html(lang="en"):
            with self.__a.head():
                self.__a.meta(charset="utf-8")
                self.__a.title(_t=self.title)
                with self.__a.style():
                    self.__a(self.style_css())
            with self.__a.body(klass=theme):
                # Iterate over markup
                for item in self.log:
                    self.__a(item)

        create_file_at_path(self.filename)
        with open(self.filename, "w") as f:
            f.write(str(self.__a))



import re

# Defines regex functions
class Regex:
    def apply(regex, text):
        new_text = text
        if regex['type'] == 'match':
            new_text = Regex.match(regex, new_text)
        elif regex['type'] == 'replace':
            new_text = Regex.replace(regex, new_text)
        return new_text
    def match(regex, text):
        find = re.search(regex['pattern'], text)
        return find.group() if find else ''
    def replace(regex, text):
        return re.sub(
            regex['old_value'],
            regex['new_value'],
            text
        )

# Defines regex Types
class RegexType:
    def match(pattern):
        return {
            'type': 'match',
            'pattern': pattern
        }
    def replace(old_value, new_value):
        return {
            'type': 'replace',
            'old_value': old_value,
            'new_value': new_value
        }